(* This is the intensionality monad discussed in the lecture notes for week 7. *)

type s = int;; (* integers model possible worlds *)
type e = char;; (* chars model individuals *)
type t = bool;; (* booleans model truth values *)

let ann = 'a';;
let bill = 'b';;
let cam = 'c';;

let left1 (x:e) = true;; (* Everyone left *)
let saw1 (x:e) (y:e) = y < x;; (* Ann saw Bill and Cam, and Bill saw Cam *)

left1 ann;;
saw1 bill ann;;
saw1 ann bill;;

(* Now we make the extension of "leave" sensitive to the world of evaluation *)
let left (x:e) (w:s) = match (x, w) with ('c', 2) -> false | _ -> true;;

left ann 1;; (* Ann left in world 1 *)
left cam 2;; (* Cam didn't leave in world 2 *)

let saw x y w = (w < 2) && (y < x);;
saw bill ann 1;; (* Ann saw Bill in world 1 *)
saw bill ann 2;; (* Ann didn't see Bill in world 2 *)

(* The intensionality reader-monad: *)
type 'a intension = s -> 'a;;
let unit x (w:s) = x;;
let bind m f (w:s) = f (m w) w;;
let lift f u v = bind u (fun x -> bind v (fun y -> f x y));;

bind (unit ann) left 1;;
bind (unit cam) left 2;;

lift saw (unit bill) (unit ann) 1;; 
lift saw (unit bill) (unit ann) 2;;

let thinks (p:s->t) (x:e) (w:s) = 
  match (x, p 2) with ('a', false) -> false | _ -> p w;;

bind (unit ann) (thinks (bind (unit bill) left)) 1;;
bind (unit ann) (thinks (bind (unit cam) left)) 1;;

